/* $Id: usb_download.c,v 1.5 2005/03/28 11:43:56 bsh Exp $ */

/*
 * Copyright (c) 2003, 2005  Genetec corporation.  All rights reserved.
 * Written by Hashimoto Kenichi for Genetec corporation.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of Genetec corporation may not be used to endorse
 *    or promote products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY GENETEC CORP. ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL GENETEC CORP.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <errno.h>
#include <err.h>
#include <strings.h>
#include <sys/stat.h>
#include <sys/stdint.h>			/* for uint32_t */

#include <dev/usb/usb.h>
#include <dev/usb/usbhid.h>

#define DOWNLOADADDRESS 0x30000000
#define	DEFAULT_USB_DEVICE	"/dev/ugen0"



void usage(void);
static int open_usb_endpoint(const char *, int, int);
static void get_endpoints_for_download(int , int *, int *);


const char *device_name = DEFAULT_USB_DEVICE;

/*
 * DNW utility downloads a program in the following format.
 *   header:
 *       download address (4-byte)
 *       program size     (4-byte)
 *   program binary
 *   checksum             (2-byte)   
 */

struct download_header {
	uint32_t  address;
	uint32_t  size;
};

int v_flag = 0;			/* verbose */

#define	verbosemsg(args...) do { if (v_flag) fprintf(stderr, args); } while(0)

int main(int argc, char *argv[])
{
	int ch;

	int i;

	FILE *fp;
	int ugen_ep0, ugen_out;
	int in_endpoint, out_endpoint;	/* endpoint address */

	char usb_dev_name[32];
	char *filename;			/* file to download */
	int download_address = DOWNLOADADDRESS;
	struct stat file_stats;
	uint32_t file_size;
	uint16_t checksum;

	struct download_header hdr;

	char *txBuf, *p;

	setprogname(argv[0]);
        while ((ch = getopt(argc, argv, "d:v")) != -1) {
                switch (ch) {
                case 'd':
                       	device_name = optarg;
                        break;
                case 'v':
                       	v_flag = 1;
                        break;
                default:
                        usage();
                        /* NOTREACHED */
                }
        }
        argc -= optind;
        argv += optind;

	if (device_name[0] != '/') 
		strcpy(usb_dev_name, "/dev/");
	else
		usb_dev_name[0] = '\0';
	strlcat(usb_dev_name, device_name, sizeof usb_dev_name);

	ugen_ep0 = open_usb_endpoint(usb_dev_name, 0, O_RDWR);

	get_endpoints_for_download(ugen_ep0, &in_endpoint, &out_endpoint);



	if (argc != 1) {
		warnx("too %s arguments", argc <= 0 ? "few" : "many");
		usage();
	}

	filename = argv[0];
	fp = fopen(filename, "r");

	if (fp == NULL)
		err(EXIT_FAILURE, "can't open %s:", filename);

	verbosemsg("file name : %s\n", filename);

	if ( stat(filename, &file_stats) < 0)
		err(EXIT_FAILURE, "failed to stat %s", filename);

	file_size = file_stats.st_size;

	verbosemsg("file size = %d\n", file_size);

	ugen_out = open_usb_endpoint(usb_dev_name, out_endpoint, O_WRONLY);

	hdr.address = download_address;
	hdr.size = file_size + 10;

	txBuf = malloc(sizeof hdr + file_size + sizeof checksum);
	if (txBuf == NULL)
		err(EXIT_FAILURE, "cant alloc");
	
	memcpy(txBuf, &hdr, sizeof hdr);
	p = txBuf + sizeof hdr;
	if (fread(p, file_size, 1, fp) != 1)
		err(EXIT_FAILURE, "read failed: %s", filename);

	checksum = 0;
	for(i = 0; i < file_size; i++){			/* attach check sum */
		checksum += (unsigned char)(p[i]);
	}

	memcpy(p+file_size, &checksum, sizeof checksum);
	write(ugen_out, txBuf, sizeof hdr + file_size + sizeof checksum);

	fclose(fp);

	close(ugen_out);

	
	verbosemsg("downloadaddress : 0x%x\n"
		   "total file size : %d\n"
		   "check sum : %d\n",
		   hdr.address, hdr.size, checksum);

	exit(EXIT_SUCCESS);
}

void
usage(void)
{
        fprintf(stderr, "Usage: %s [-d device_name] [-v] filename\n", getprogname());
        exit(EXIT_FAILURE);
        /* NOTREACHED */
}

static int
open_usb_endpoint(const char *usb_device_name, int endpoint_number, int mode)
{
	int fd;
	char name[100];

	snprintf(name, sizeof name,"%s.%02d", usb_device_name, endpoint_number);
        if((fd = open(name, mode)) < 0)
                err(EXIT_FAILURE, "failed to open %s", name);

	return fd;
}

static void
get_endpoints_for_download(int ugen_ep0, int *in, int *out)
{
	int in_endpoint = 0;
	int out_endpoint = 0;
	int i, j, k;
	struct usb_device_info udi;
	struct usb_config_desc ucd;
	struct usb_interface_desc uid;
	struct usb_endpoint_desc ued;

	if(ioctl(ugen_ep0, USB_GET_DEVICEINFO, &udi) < 0)
		err(EXIT_FAILURE, "USB_GET_DEVICEINFO");

	verbosemsg("addr :%d\n", udi.udi_addr);
	verbosemsg("bNumConfigurations :%d\n", udi.udi_config);

	for (i=0; i<udi.udi_config; i++) {
		verbosemsg("config_index :%d\n", i);
		ucd.ucd_config_index = i;
		if(ioctl(ugen_ep0, USB_GET_CONFIG_DESC, &ucd) < 0)
			err(EXIT_FAILURE, "USB_GET_CONFIG_DESC");

		verbosemsg("bNumInterface :%d\n", ucd.ucd_desc.bNumInterface);

		for (j = 0; j < ucd.ucd_desc.bNumInterface; j++) {
			if(v_flag)
				fprintf(stderr, "interface_indeex :%d\n", j);
			uid.uid_interface_index = j;
			uid.uid_alt_index = 0;
			if(ioctl(ugen_ep0, USB_GET_INTERFACE_DESC, &uid) < 0)
				err(EXIT_FAILURE, "USB_GET_INTERFACE_DESC");

			verbosemsg("bAlternateSetting :%d\n", uid.uid_desc.bAlternateSetting);
			verbosemsg("bNumEndpoints :%d\n", uid.uid_desc.bNumEndpoints);

			for (k = 0; k < uid.uid_desc.bNumEndpoints; k++) {
				ued.ued_config_index = i;
				ued.ued_interface_index = j;
				ued.ued_alt_index = uid.uid_desc.bAlternateSetting;
				ued.ued_endpoint_index = k;
				if(ioctl(ugen_ep0, USB_GET_ENDPOINT_DESC, &ued) < 0)
					err(EXIT_FAILURE, "USB_GET_ENDPOINT_DESC");

				if((0x80 & ued.ued_desc.bEndpointAddress) == 0x80) {
					in_endpoint = (ued.ued_desc.bEndpointAddress & 0x07);
					verbosemsg("in_endpoint :%x\n", in_endpoint);
				} else {
					out_endpoint = (ued.ued_desc.bEndpointAddress & 0x07);
					verbosemsg("out_endpoint :%x\n", out_endpoint);
				}
			}
		}
	}


	if (in_endpoint <= 0)
		errx(EXIT_FAILURE, "%s: can't find endpoint for %s", device_name, "input");
	if (out_endpoint <= 0)
		errx(EXIT_FAILURE, "%s: can't find endpoint for %s", device_name, "output");

	*in = in_endpoint;
	*out = out_endpoint;
}
